package knapsack;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;
import java.util.StringTokenizer;
import knapsack.algorithms.*;

/**
 *
 * @author Jaroslav Fibichr
 */
public class KnapsackSolver {

    private static final int MIN_TIME_THRESHOLD = 100;
    List<Instance> instances = new ArrayList<Instance>();
    String fileName;
    Algorithm algorithm;

    public KnapsackSolver(String fileName) {
        this.fileName = fileName;
    }

    public void loadData() {
        String line;
        int num;
        FileReader fr;
        try {
            fr = new FileReader(fileName);
            BufferedReader buffIn = new BufferedReader(fr);

            while ((line = buffIn.readLine()) != null) {
                Instance instance = new Instance();
                StringTokenizer st = new StringTokenizer(line);

                instance.id = st.nextToken();
                instance.itemCount = Integer.parseInt(st.nextToken());
                instance.capacity = Integer.parseInt(st.nextToken());

                instance.items = new Item[instance.itemCount];
                instance.remValues = new int[instance.itemCount];

                // load items
                for (int i = 0; st.hasMoreTokens(); i++) {
                    Item item = new Item();
                    item.weight = Integer.parseInt(st.nextToken());
                    item.value = Integer.parseInt(st.nextToken());
                    item.seq = i; // original sequence number (reference after sorting)
                    instance.items[i] = item;
                    instance.totalValue += item.value;
                    instance.remValues[i] = instance.totalValue;
                }
                instance.solMatrix = new Solution[instance.itemCount][instance.capacity + 1];
                instances.add(instance);
            }

            fr.close();
        } catch (FileNotFoundException ex) {
            System.out.println("File open error.");
        } catch (IOException ex) {
            System.out.println("File read error.");
        }
    }

    /**
     * @param args the command line arguments
     */
    public static void main(String[] args) {

        // time counter
        long start;
        double elapsed;

        if (args.length < 6) {
            System.out.println("Not enough parameters.");
            System.exit(0);
        }
        KnapsackSolver kss = new KnapsackSolver(args[0]);

        // get Genetic algo parameters
        Genetic.POPULATION_SIZE = Integer.parseInt(args[1]);
        Genetic.GENERATION_COUNT = Integer.parseInt(args[2]);
        Genetic.SELECTION_PRESSURE = Double.parseDouble(args[3]);
        Genetic.P_CROSSOVER = Double.parseDouble(args[4]);
        Genetic.P_MUTATION = Double.parseDouble(args[5]);
        
        // load all instances data (whole file)
        kss.loadData();

        Algorithm[] algos = {new Dynamic(), new Genetic()}; // first must give optimal values!
        Result[] results = new Result[algos.length];
        for (int i = 0; i < algos.length; i++) {
            results[i] = new Result();
        }

        // run all instances
        for (int i = 0; i < kss.instances.size(); ++i) {

            Instance ins = kss.instances.get(i);

//            System.out.print("Instance: ");
//            System.out.println(ins.id + " " + ins.itemCount + " ");

            for (int j = 0; j < algos.length; j++) {
                kss.algorithm = algos[j];
                Result result = results[j];

                start = System.nanoTime();
                Solution solution = new Solution(ins.capacity, ins.itemCount);
                solution = kss.algorithm.run(ins, solution);
                elapsed = ((double) (System.nanoTime() - start)) / 1000000;

                // too short ( < 1000 ms )
                if (elapsed < MIN_TIME_THRESHOLD) {
                    long its = Math.round(MIN_TIME_THRESHOLD / elapsed);   // repeat n-times to fill 100 ms
                    start = System.nanoTime();
                    for (int k = 0; k < its; k++) {
                        solution = new Solution(ins.capacity, ins.itemCount);
                        solution = kss.algorithm.run(ins, solution);
                    }
                    result.iterations = its;
                    elapsed = ((double) (System.nanoTime() - start)) / 1000000 / its;
                }

                result.totalTime += elapsed;
                result.value = solution.value;
//                System.out.print(" > " + kss.algorithm.getName() + ": " + elapsed + " ms ");
//                solution.print();

                // error evaluation
                if (j != 0) {
                    double error = ((double) Math.abs(results[0].value - results[j].value)) / results[0].value;
//                    maxError = error > maxError ? error : maxError;
                    result.totalError += error;
                }
            } // algorithm


        } // instance

//        System.out.println("Overall results:");
//        for (int i = 0; i < algos.length; i++) {
//            System.out.println(" >> Time " + algos[i].getName() + ": \t" + results[i].totalTime / kss.instances.size());
//        }
//        for (int i = 0; i < algos.length; i++) {
//            System.out.println(" >> Error " + algos[i].getName() + ":\t" + results[i].totalError / kss.instances.size());
//        }
//        for (int i = 0; i < algos.length; i++) {
//            String time = String.format("%1$e", results[i].totalTime / kss.instances.size());
//            System.out.print(" >> " + algos[i].getName() + ":      \t" + time);
//            System.out.print(" \t " + results[i].iterations);
//            System.out.println(" \t " + results[i].totalError / kss.instances.size());
//        }

        // print in table (to put data into Excel table)
        String names = "";
        String values = "";
        String iters = "";
        for (int i = 0; i < algos.length; i++) {
            names += padLeft(algos[i].getName());
            iters += padLeft(results[i].iterations+"");
            values += padLeft(String.format("%1.2e", results[i].totalTime / kss.instances.size()));
        }
        names += padLeft("H_err") + padLeft("FPTAS_err");
        values += padLeft(String.format("%1.2f%%",results[1].totalError / kss.instances.size() * 100));
//        values += padLeft(String.format("%1.2f%%",results[4].totalError / kss.instances.size() * 100));
        values = values.replace(".", ",");
//        System.out.println(names);
        System.out.println(values);
//        System.out.println(iters);


    } // main

    public static String padLeft(String s) {
//        return String.format("%" + 14 + "s", s);
        return s + "\t";
    }
}
